.SYNTAX Program

PrimaryExpression =
  PrimaryExpressionHead $PrimaryExpressionTail;

PrimaryExpressionHead =
  '(' Expression ')' /
  'function' FunctionRest /
  'new' .PRE['new'] .ID .OUT(*) ('(' ArgumentList ')' / .EMPTY) .CLOSE /
  .ID .OUT(*) /
  .NUMBER .PRE['quote'] .OUT(*) .CLOSE /
  .STRING .PRE['quote'] .OUT(*) .CLOSE;

PrimaryExpressionTail =
  '[' .POST['array-access'] Expression ']' .CLOSE /
  '.' .POST['field-access'] .ID .PRE['quote'] .OUT(*) .CLOSE .CLOSE
    ('(' .POST['jsend'] .DESTRUCT ArgumentList ')' .CLOSE /
     '=' ('function' .POST['store-method'] .DESTRUCT FunctionRest .CLOSE /
     .POST['store-value'] .DESTRUCT Expression .CLOSE) /
     .EMPTY) /
  '(' .POST['function-call'] ArgumentList ')' .CLOSE;

ArgumentList =
  Expression $(',' Expression) /
  .EMPTY;

FunctionRest =
  .PRE['lambda'] '(' Formals ')' '{' SourceElements '}' .CLOSE;

Formals =
  .PRE['_closure'] .OUT('this') (.ID .OUT(*) $(',' .ID .OUT(*)) / .EMPTY) .CLOSE;

PostfixExpression =
  PrimaryExpression ('++' .POST['incr'] .CLOSE /
                     '--' .POST['decr'] .CLOSE /
                     .EMPTY);

UnaryExpression =
  '+' UnaryExpression /
  '-' .PRE['unary'] .OUT('negated') UnaryExpression .CLOSE /
  '!' .PRE['unary'] .OUT('not')     UnaryExpression .CLOSE /
  PostfixExpression;

MultiplicativeExpression =
  UnaryExpression $('*' .POST['binary'] .OUT('*')  UnaryExpression .CLOSE /
                    '/' .POST['binary'] .OUT('//') UnaryExpression .CLOSE /
                    '%' .POST['binary'] .OUT('\\') UnaryExpression .CLOSE);

AdditiveExpression =
  MultiplicativeExpression $('+' .POST['binary'] .OUT('+') MultiplicativeExpression .CLOSE /
                             '-' .POST['binary'] .OUT('-') MultiplicativeExpression .CLOSE);

RelationalExpression =
  AdditiveExpression $('<=' .POST['binary'] .OUT('<=') AdditiveExpression .CLOSE /
                       '<'  .POST['binary'] .OUT('<')  AdditiveExpression .CLOSE /
                       '>=' .POST['binary'] .OUT('>=') AdditiveExpression .CLOSE /
                       '>'  .POST['binary'] .OUT('>')  AdditiveExpression .CLOSE);

EqualityExpression =
  RelationalExpression $('==' .POST['binary'] .OUT('=')  RelationalExpression .CLOSE /
                         '!=' .POST['binary'] .OUT('~~') RelationalExpression .CLOSE);

LogicalANDExpression =
  EqualityExpression $('&&' .POST['binary'] .OUT('&&') EqualityExpression .CLOSE);

LogicalORExpression =
  LogicalANDExpression $('||' .POST['binary'] .OUT('||') LogicalANDExpression .CLOSE);

Expression =
  LogicalORExpression ('?' .POST['if'] Expression ':' Expression .CLOSE /
                       '=' .POST['set'] Expression .CLOSE /
                       .EMPTY);

Statement =
  '{' .PRE['begin'] (Statement $Statement / .EMPTY) .CLOSE '}' /
  VariableDeclaration /
  'if' .PRE['if'] '(' Expression ')' Statement ('else' Statement / .EMPTY) .CLOSE /
  'while' .PRE['while'] '(' Expression ')' Statement .CLOSE /
  'do' .PRE['do'] Statement 'while' '(' Expression ')' .CLOSE ';' /
  'for' .PRE['for'] '(' VariableDeclaration (Expression / .EMPTY .OUT('1')) ';' (Expression / .OUT('0')) ')' Statement .CLOSE /
  'return' .PRE['return'] (Expression / .EMPTY) ';' .CLOSE /
  Expression ';';

VariableDeclaration =
  'var' VariableBinding $(',' VariableBinding) ';';

VariableBinding =
  .ID .PRE['declare'] .OUT(*) ('=' Expression / .EMPTY .OUT('null')) .CLOSE;

SourceElement =
 'function' .PRE['declare'] .ID .OUT(*) FunctionRest .CLOSE /
  Statement;

SourceElements =
  SourceElement $SourceElement /
  .EMPTY;

Program = .PRE['prog'] SourceElements .CLOSE;

.END

