{ import: st80 }
{ import: File }
{ import: serialise }

Character isWhite               [^self <= $ ]
Character isLetterOrDigit       [^self isLetter or: [self isDigit]]

MetaIVMachine : Object ( input pos oldPos deleted deletedIsNum deletedIsIdentifier switch recording debug tokens )

MetaIVMachine initialize        [debug := false]
MetaIVMachine debug: d          [debug := d]

MetaIVMachine try               [oldPos := pos. switch := true]
MetaIVMachine startRecording    [deleted := OrderedCollection new. recording := true]
MetaIVMachine stopRecording     [recording := false]
MetaIVMachine fail              [pos := oldPos. switch := false. self stopRecording]

MetaIVMachine atEnd             [^pos > input size]
MetaIVMachine hasNext           [^self atEnd not]
MetaIVMachine peek              [^input at: pos]
MetaIVMachine peekIs: c         [^self hasNext and: [self peek == c]]
MetaIVMachine peekSatisfies: p  [^self hasNext and: [p value: self peek]]

MetaIVMachine error: msg
[
    | line col |
    line := 1.
    col := 1.
    1 to: pos - 1 do: [:idx | (input at: idx) = $\n ifTrue:  [line := line + 1. col := 1] ifFalse: [
			      (input at: idx) = $\r ifFalse: [col := col + 1]]].
    ('error on line ', line printString, ', column ', col printString, ': ') put.
    super error: msg
]

MetaIVMachine next
[
    | r |
    r := self peek.
    recording ifTrue: [deleted addLast: r].
    pos := pos + 1.
    ^r.
]

MetaIVMachine skipWhite         [[self peekSatisfies: [:c | c isWhite]] whileTrue: [self next]]

MetaIVMachine seqFirst: firstp rest: restp record: record
[
    self skipWhite; try.
    record ifTrue: [self startRecording].
    (self peekSatisfies: firstp) ifTrue:  [self next.
					   [self peekSatisfies: restp] whileTrue: [self next]]
				 ifFalse: [self fail].
    record ifTrue: [self stopRecording].
]

OrderedCollection label1        [^self second]
OrderedCollection label1: l     [self at: 2 put: ('L', l printString) asSymbol]
OrderedCollection label2        [^self first]
OrderedCollection label2: l     [self at: 1 put: ('L', l printString) asSymbol]

MetaIVMachine tst: s
[
    | allLettersAndDigits old |
    self skipWhite; try.
    s do: [:c | (self peekIs: c) ifTrue:  [self next]
				 ifFalse: [^self fail]].
    " here is a little hack to prevent identifiers that start with a keyword like 'document' work "
    allLettersAndDigits := true.
    s do: [:c | c isLetterOrDigit ifFalse: [allLettersAndDigits := false]].
    (allLettersAndDigits and: [self peekSatisfies: [:p | p isWhite not and: [p isLetterOrDigit]]])
	ifTrue: [^self fail].
    " this makes sure that we don't match on '+' when it should be '++', etc. "
    old := pos.
    tokens do: [:t |
		pos := old.
		(t size > s size and: [t beginsWith: s])
		    ifTrue: [| match |
			     match := true.
			     s size + 1 to: t size do:
				[:i | (self peekIs: (t at: i)) ifTrue: [self next]
							       ifFalse: [match := false]].
			     match ifTrue: [^self fail]]]]

MetaIVMachine sr
[
    self skipWhite; try; startRecording.
    (self peekIs: $') ifTrue:  [self next.
				deleted removeLast. 
				[(self peekIs: $') not] whileTrue: [self atEnd ifTrue: [^self fail].
								    self next].
				self stopRecording.
				deletedIsNum := deletedIsIdentifier := false.
				self next]
		      ifFalse: [self fail]
]

MetaIVMachine indicesOfInstructionsIn: code
[
    | r |
    r := IdentitySet with: 1.
    2 to: code size do:
	[:i |
	 ((r includes: i - 1) and: [#(TST CLL B BT BF CL LB PRE PST LB) includes: (code at: i - 1)]) not
	     ifTrue: [r add: i]].
    ^r
]

MetaIVMachine resolveLabelsIn: code withInstructionsAt: instructions
[
    | addresses |
    code := OrderedCollection withAll: code.
    addresses := Dictionary new.
    instructions do: [:idx | (code at: idx) = #LB ifTrue: [addresses at: (code at: idx + 1) put: idx + 2]].
    instructions do:
	[:idx |
	 (#(CLL B BT BF) includes: (code at: idx)) ifTrue: 
	     [code at: idx + 1
		   put: (addresses at: (code at: idx + 1)
		   ifAbsent: [Smalltalk error: 'undefined label ', (code at: idx + 1)])]].
    ^code
]

MetaIVMachine interp: code withInput: ipt
[
    | pc stack output labelCount dispatcher outputStack instructions |
    instructions := self indicesOfInstructionsIn: code.
    code := self resolveLabelsIn: code withInstructionsAt: instructions.
    tokens := IdentitySet new.
    instructions do: [:idx | (code at: idx) = #TST ifTrue: [tokens add: (code at: idx + 1)]].
    input := ipt.
    pos := 1.
    switch := false.
    recording := false.
    pc := 1.
    stack := OrderedCollection new.
    stack addFirst: code size; " the address of the END instruction "
          addFirst: nil;       " the first label  "
	  addFirst: nil.       " the second label "
    output := OrderedCollection new.
    outputStack := OrderedCollection new.
    outputStack add: output.
    labelCount := 0.
    dispatcher := IdentityDictionary new.
    dispatcher
	at: #TST put: [| s |
		       s := code at: pc.
		       pc := pc + 1.
		       self tst: s];
	at: #ID  put: [self seqFirst: [:c | c isLetter] rest: [:c | c isLetterOrDigit] record: true.
		       deletedIsIdentifier := true.
		       deletedIsNum := false];
"       this old version does not know how to deal with + and -
	at: #NUM put: [self seqFirst: [:c | c isDigit]  rest: [:c | c isDigit]	       record: true.
		       deletedIsIdentifier := false.
		       deletedIsNum := true];
"
	at: #NUM put: [| p neg |
		       p := pos.
		       neg := false.
		       (self peekIs: $~) ifTrue: [self next. neg := true].
		       self seqFirst: [:c | c isDigit] rest:  [:c | c isDigit]         record: true.
		       switch ifTrue:  [neg ifTrue: [deleted addFirst: $~].
					deletedIsIdentifier := false.
					deletedIsNum := true]
			      ifFalse: [pos := p]];
        at: #SR  put: [self sr];
	at: #CLL put: [| addr |
		       addr := code at: pc.
		       pc := pc + 1.
		       stack addFirst: pc;  " the exit address "
			     addFirst: nil; " the first label  "
			     addFirst: nil. " the second label "
		       pc := addr];
	at: #R   put: [stack removeFirst; " pop the first label "
			     removeFirst. " pop the second label "
		       pc := stack removeFirst];
	at: #SET put: [switch := true];
	at: #B   put: [pc := code at: pc];
	at: #BT  put: [switch ifTrue:  [pc := code at: pc]
			      ifFalse: [pc := pc + 1]];
	at: #BF  put: [switch ifFalse: [pc := code at: pc]
			      ifTrue:  [pc := pc + 1]];
	at: #BE  put: [switch ifFalse: [self error: 'parse error']];
	at: #CL  put: [| s |
		       s := code at: pc.
		       pc := pc + 1.
		       output add: s asSymbol];
	at: #CI  put: [switch ifTrue:  [deletedIsNum
					    ifTrue:  [| n neg |
						      n := 0.
						      deleted first = $~
							ifTrue:  [neg := true. deleted removeFirst]
							ifFalse: [neg := false].
						      deleted do: [:c | n := n * 10 + c asciiValue - $0 asciiValue].
						      output add: (neg ifFalse: [n] ifTrue: [n negated])]
					    ifFalse: [| s |
						      s := String new: deleted size.
						      deleted doWithIndex: [:c :idx | s at: idx put: c].
						      output add: (deletedIsIdentifier ifTrue:  [s asSymbol]
										       ifFalse: [s])]]
			      ifFalse: [self error: '(INTERNAL) cannot execute CI when switch=false']];
	at: #GN1 put: [stack label1 = nil ifTrue: [stack label1: (labelCount := labelCount + 1)].
		       output add: stack label1];
	at: #GN2 put: [stack label2 = nil ifTrue: [stack label2: (labelCount := labelCount + 1)].
		       output add: stack label1];
	at: #END put: [self skipWhite.
		       self atEnd ifTrue:  [^output]
				  ifFalse: [self error: 'end-of-file expected']];
	at: #LB  put: [pc := pc + 1];
        at: #NIL put: [output add: nil];
        at: #SWP put: [| x y |
		       x := output removeLast.
		       y := output removeLast.
		       output add: x; add: y];
	at: #DUP put: [output add: output last];
	at: #DST put: [| expr |
		       expr := output removeLast.
		       2 to: expr size do: [:idx | output add: (expr at: idx)]];
	at: #CLS put: [outputStack removeLast.
		       output := outputStack last];
	at: #PR  put: [output add: OrderedCollection new.
		       output := output last.
		       outputStack add: output];
	at: #PRE put: [| type |
		       type := code at: pc.
		       pc := pc + 1.
		       output add: OrderedCollection new.
		       output := output last.
		       outputStack add: output.
		       output add: type asSymbol];
	at: #PS  put: [| last |
		       last := output removeLast.
		       output add: OrderedCollection new.
		       output := output last.
		       outputStack add: output.
		       output add: last];
	at: #PST put: [| type last |
		       type := code at: pc.
		       pc := pc + 1.
		       last := output removeLast.
		       output add: OrderedCollection new.
		       output := output last.
		       outputStack add: output.
		       output add: type asSymbol; add: last].
    [true]
	whileTrue:
	    [| insn arg |
	     insn := code at: pc.
	     pc := pc + 1.
	     debug ifTrue: [('SWITCH ', (switch ifTrue: ['ON'] ifFalse: ['OFF'])) putln.
			    'STACK ' put.
			    stack println.
			    ('CODE ', (pc - 1) printString, ': ', insn printString) put.
			    pc <= code size ifTrue: [(', ', (code at: pc) printString) put].
			    pc + 1 <= code size ifTrue: [', ...' put].
			    '' putln.
			    ('INPUT ', pos printString, ': ') put.
			    0 to: 29 do: [:i | pos + i < input size ifTrue: [| c |
									     c := input at: pos + i.
									     c == $\n ifTrue: [c := '{newline}'].
									     c put]].
			    '' putln].
	     (dispatcher at: insn ifAbsent: [self error: '(INTERNAL) invalid instruction ', insn printString]) value.
	     debug ifTrue: [output println. '' putln]].
]

Object arrayify
[
    ^self
]

OrderedCollection arrayify
[
    self doWithIndex: [:e :idx | self at: idx put: e arrayify].
    ^self asArray
]

[
    | out |
    Smalltalk arguments size ~~ 2 ifTrue: ['usage: m4 compilerFile inputFile' putln. Smalltalk quit].
    (MetaIVMachine new
	interp:    (File contentsOf: Smalltalk arguments first) readStream load
	withInput: (File contentsOf: Smalltalk arguments second)) arrayify println.
]

