{ import: Object-imp }
{ import: Object-lib }

Link : Object ( prevLink nextLink onList )

Link prevLink		[ ^prevLink ]
Link prevLink: aLink	[ ^prevLink := aLink ]
Link nextLink		[ ^nextLink ]
Link nextLink: aLink	[ ^nextLink := aLink ]
Link onList		[ ^onList ]
Link onList: aList	[ ^onList := aList ]

Link offList		[  onList := nil ]

Link unlink
[
    prevLink nextLink: nextLink.
    nextLink prevLink: prevLink.
    self offList.
]


LinkedList : Object ( root )

LinkedList new
[
    self := super new.
    root := Link new.
    root prevLink: root.
    root nextLink: root.
]

LinkedList isEmpty	[ ^root nextLink == root ]
LinkedList notEmpty	[ ^root nextLink ~~ root ]

LinkedList emptyCheck	[ root nextLink == root ifTrue: [self errorEmptyCollection] ]

LinkedList addFirst: aLink
[
    | prev |
    aLink onList notNil ifTrue: [self error: 'link is already on a list: ', aLink printString].
    aLink prevLink: root.
    aLink nextLink: root nextLink.
    root nextLink prevLink: aLink.
    root nextLink: aLink.
    aLink onList: self.
    ^aLink.
]

LinkedList addLast: aLink
[
    ^self add: aLink
]

LinkedList add: aLink
[
    | prev |
    aLink onList notNil ifTrue: [self error: 'link is already on a list: ', aLink printString].
    aLink nextLink: root.
    aLink prevLink: root prevLink.
    root prevLink nextLink: aLink.
    root prevLink: aLink.
    aLink onList: self.
    ^aLink.
]

LinkedList removeFirst
[
    | link |
    self emptyCheck.
    link := root nextLink.
    (root nextLink: link nextLink) prevLink: root.
    ^link offList
]

LinkedList removeLast
[
    | link |
    self emptyCheck.
    link := root prevLink.
    (root prevLink: link prevLink) nextLink: root.
    ^link offList
]

LinkedList first
[
    self emptyCheck.
    ^root nextLink
]

LinkedList last
[
    self emptyCheck.
    ^root prevLink
]

LinkedList size
[
    | count aLink |
    count := 0.
    aLink := root.
    [(aLink := aLink nextLink) == root] whileFalse: [count := 1 + count].
    ^count
]

LinkedList do: unaryBlock
[
    | aLink |
    aLink := root.
    [(aLink := aLink nextLink) == root] whileFalse: [unaryBlock value: aLink]
]

LinkedList reverseDo: unaryBlock
[
    | aLink |
    aLink := root.
    [(aLink := aLink prevLink) == root] whileFalse: [unaryBlock value: aLink]
]

LinkedList from: aLink do: unaryBlock
[
    [aLink == root]
	whileFalse:
	   [unaryBlock value: aLink.
	    aLink := aLink nextLink]
]

"----------------------------------------------------------------"

LineScanner : Object ( input line lineNumber )

LineScanner on: aStringOrStream
[
    self := self new initialize.
    input := aStringOrStream readStream.
]

LineScanner initialize
[
    line := WriteStream on: (String new: 256).
    lineNumber := 0.
]

" 5.1.1.2  Translation phases

  1. Physical source file multibyte characters are mapped, in an implementation-
     defined manner, to the source character set (introducing new-line characters for
     end-of-line indicators) if necessary.  Trigraph sequences are replaced by 
     corresponding single-character internal representations. "

LineScanner discardEOL
[
    | c |
    line nextPut: $\n; reset.
    lineNumber := lineNumber + 1.
    c := input next.
    c == $\n
	ifTrue:  [[input peek == $\r] whileTrue: [input next]]
	ifFalse:
	   [c == $\n
		ifTrue:  [[input peek == $\r] whileTrue: [input next]]
		ifFalse: [self error: 'no EOL']]
]

" 2. Each instance of a backslash character (\) immediately followed by a new-line 
     character is deleted, splicing physical source lines to form logical source lines. 
     Only the last backslash on any physical source line shall be eligible for being part 
     of such a splice.  A source file that is not empty shall end in a new-line character, 
     which shall not be immediately preceded by a backslash character before any such 
     splicing takes place. "

LineScanner refill
[
    | c |
    line reset.
    input atEnd ifTrue: [^line reset].
    [(c := input peek) notNil]
	whileTrue:
	   [c == $\\
		ifTrue:
		   [input next.
		    (input peek == $\n or: [input peek == $\r])
			ifTrue:  [^self discardEOL]
			ifFalse: [line nextPut: c]]
		ifFalse:
		   [(c == $\n or: [c == $\r])
			ifTrue:  [^self discardEOL]
			ifFalse: [line nextPut: input next]]]
]

LineScanner next
[
    line atEnd ifTrue: [self refill].
    ^line atEnd ifFalse: [line next].
]

LineScanner peek
[
    line atEnd ifTrue: [self refill].
    ^line atEnd ifFalse: [line peek].
]

LineScanner match: aCollection
[
    0 to: aCollection size - 1 do: [:i | (line peek: i) == (aCollection at: i) ifFalse: [^false]].
    ^true
]

LineScanner skip: aCollection
[
    (self match: aCollection) ifFalse: [^false].
    line skip: aCollection size.
    ^true
]

"----------------------------------------------------------------"

C99TokenList : LinkedList ()

C99TokenList removeOuterBlanks
[
    [self notEmpty and: [root nextLink isBlank]] whileTrue: [self removeFirst].
    [self notEmpty and: [root prevLink isBlank]] whileTrue: [self removeLast].
]

C99TokenList removeBlanks
[
    self do: [:tok | tok isBlank ifTrue: [tok unlink]].		"iteration is stable under deletion"
]

C99TokenList combineBlanks
[
    | prevBlank |
    prevBlank := false.
    self do: [:tok |
	tok isBlank
	    ifTrue:
	       [prevBlank
		    ifTrue:  [tok unlink]
		    ifFalse: [tok text: ' '].
		prevBlank := true]
	    ifFalse:
	       [prevBlank := false]]
]

"----------------------------------------------------------------"

C99Token : Link ( text note line )

C99Token withText: textString line: lineNumber note: aNote
[
    self := self withText: textString line: lineNumber.
    note := aNote.
]

C99Token withText: textString line: lineNumber
[
    self := self new.
    text := textString.
    line := lineNumber.
]

C99Token copy	[ ^self withText: text line: line note: note ]

C99Token text		[ ^text ]
C99Token text: aString	[ ^text := aString ]
C99Token note		[ ^note ]
C99Token note: aNote	[ ^note := aNote ]
C99Token line		[ ^line ]
C99Token line: aNumber	[ ^line := aNumber ]

C99Token isEOL		[ ^note == #eol ]
C99Token notEOL		[ ^note ~~ #eol ]
C99Token isComment	[ ^note == #comment ]

C99Token error: message	[ super error: line printString, ': ERROR: ', message ]
C99Token warn:  message	[ StdErr nextPutAll: line printString, ': WARNING: ', message; cr ]

C99Token printOn: aStream
[
    aStream nextPut: ${; nextPutAll: self debugName; space; print: note; tab; print: text; nextPut: $}
]

C99Token putTextOn: aStream
[
    "self isComment ifFalse: ["aStream nextPutAll: text"]"
]



C99Directive : C99Token ( name body )

C99Directive new
[
    self := super new.
    body := C99TokenList new.
]

C99Directive name: aToken	[ ^name := aToken ]
C99Directive body		[ ^body ]
C99Directive body: aToken	[ ^body := aToken ]

C99Directive printOn: aStream
[
    super printOn: aStream.
    body do: [:tok | tok printOn: aStream].
]

C99Directive putTextOn: aStream
[
    aStream nextPut: $#.
    body do: [:tok | tok putTextOn: aStream].
    aStream cr.
]

C99GroupSection : C99Directive ( nextGroup )

C99GroupSection nextGroup		[ ^nextGroup ]
C99GroupSection nextGroup: aToken	[ ^nextGroup := aToken ]

C99GroupDirective : C99GroupSection ( enclosingGroup )

C99GroupDirective enclosingGroup		[ ^enclosingGroup ]
C99GroupDirective enclosingGroup: aToken	[ ^enclosingGroup := aToken ]

C99DirectiveIf  : C99GroupDirective ()
  C99DirectiveIfdef : C99DirectiveIf ()
    C99DirectiveIfndef : C99DirectiveIfdef ()

C99DirectiveElif : C99GroupDirective ()

C99DirectiveElse : C99GroupDirective ()

C99Directive     isDirectiveElse	[ ^false ]
C99DirectiveElse isDirectiveElse	[ ^true ]

C99DirectiveEndif : C99Directive ()

C99Directive      isDirectiveEndif	[ ^false ]
C99DirectiveEndif isDirectiveEndif	[ ^true ]

C99DirectiveInclude : C99Directive ()

C99DirectiveDefine : C99Directive ()

C99DirectiveUndef : C99Directive ()

C99DirectiveLine : C99Directive ()

C99DirectiveError : C99Directive ()

C99DirectivePragma : C99Directive ()

C99Directive	    doDirectiveIn: anObject	[ ^self error: self debugName, ' should have overridden doDirectiveIn:' ]

C99DirectiveIf	    doDirectiveIn: anObject	[ ^anObject doDirectiveIf:      self ]
C99DirectiveIfdef   doDirectiveIn: anObject	[ ^anObject doDirectiveIfdef:   self ]
C99DirectiveIfndef  doDirectiveIn: anObject	[ ^anObject doDirectiveIfndef:  self ]
C99DirectiveElif    doDirectiveIn: anObject	[ ^anObject doDirectiveElif:    self ]
C99DirectiveElse    doDirectiveIn: anObject	[ ^anObject doDirectiveElse:    self ]
C99DirectiveEndif   doDirectiveIn: anObject	[ ^anObject doDirectiveEndif:   self ]
C99DirectiveInclude doDirectiveIn: anObject	[ ^anObject doDirectiveInclude: self ]
C99DirectiveDefine  doDirectiveIn: anObject	[ ^anObject doDirectiveDefine:  self ]
C99DirectiveUndef   doDirectiveIn: anObject	[ ^anObject doDirectiveUndef:   self ]
C99DirectiveLine    doDirectiveIn: anObject	[ ^anObject doDirectiveLine:    self ]
C99DirectiveError   doDirectiveIn: anObject	[ ^anObject doDirectiveError:   self ]
C99DirectivePragma  doDirectiveIn: anObject	[ ^anObject doDirectivePragma:  self ]

C99DivideInto : C99Token ()
C99Divide : C99Token ()
C99Paste : C99Token ()
C99Hash : C99Token ()
C99LeftShiftInto : C99Token ()
C99LeftShift : C99Token ()
C99LessEqual : C99Token ()
C99Begin : C99Token ()
C99OpenParen : C99Token ()		C99OpenParen isOpenParen	[ ^true ]	C99Token isOpenParen	[ ^false ]
C99OpenBracket : C99Token ()
C99OpenBrace : C99Token ()
C99Less : C99Token ()
C99RightShiftInto : C99Token ()
C99RightShift : C99Token ()
C99GreaterEqual : C99Token ()
C99Greater : C99Token ()
C99LogAnd : C99Token ()
C99AndInto : C99Token ()
C99And : C99Token ()
C99LogOr : C99Token ()
C99OrInto : C99Token ()
C99Or : C99Token ()
C99MultiplyInto : C99Token ()
C99Multiply : C99Token ()
C99Decrement : C99Token ()
C99SubtractInto : C99Token ()
C99Member : C99Token ()
C99MemberRef : C99Token ()
C99Subtract : C99Token ()
C99Increment : C99Token ()
C99AddInto : C99Token ()
C99Add : C99Token ()
C99NotEqual : C99Token ()
C99Not : C99Token ()
C99CloseParen : C99Token ()
C99CloseBracket : C99Token ()
C99CloseBrace : C99Token ()
C99Colon : C99Token ()
C99Equal : C99Token ()
C99Assign : C99Token ()
C99Paste : C99Token ()
C99Hash : C99Token ()
C99End : C99Token ()
C99ModulusInto : C99Token ()
C99Modulus : C99Token ()
C99XorInto : C99Token ()
C99Xor : C99Token ()
C99Ellipsis : C99Token ()
C99Comma : C99Token ()
C99Semicolon : C99Token ()
C99Query : C99Token ()
C99Complement : C99Token ()

C99Other : C99Token ()
C99Number : C99Token ()
C99Character : C99Token ()
C99String : C99Token ()
C99Identifier : C99Token ()
C99Blank : C99Token ()

C99Token isIdentifier	[ ^false ]		C99Identifier isIdentifier	[ ^true ]
C99Token notIdentifier	[ ^true ]		C99Identifier notIdentifier	[ ^false ]

C99Token isBlank	[ ^false ]		C99Blank isBlank	[ ^true ]

C99Character putTextOn: aStream	[ aStream nextPut: $'; nextPutAll: text; nextPut: $' ]
C99String    putTextOn: aStream	[ aStream nextPut: $"; nextPutAll: text; nextPut: $" ]

C99Scanner : LineScanner ( thisChar characterTable directiveTable stringTable text bol activeGroup tokenLineNumber )

C99Scanner initialize
[
    super initialize.
    (characterTable := Array new: 256 withAll: #xIllegal)
	atAll: '!'	 put: #xPling;
	atAll: '#'	 put: #xHash;
	atAll: '%'	 put: #xPercent;
	atAll: '&'	 put: #xAmpersand;
	atAll: '('	 put: #xLparen;
	atAll: ')'	 put: #xRparen;
	atAll: '*'	 put: #xStar;
	atAll: '+'	 put: #xPlus;
	atAll: ','	 put: #xComma;
	atAll: '-'	 put: #xMinus;
	atAll: '.'	 put: #xDot;
	atAll: '/'	 put: #xSlash;
	atAll: ':'	 put: #xColon;
	atAll: ';'	 put: #xSemicolon;
	atAll: '<'	 put: #xLangle;
	atAll: '='	 put: #xEqual;
	atAll: '>'	 put: #xRangle;
	atAll: '?'	 put: #xQuery;
	atAll: '['	 put: #xLbracket;
	atAll: ']'	 put: #xRbracket;
	atAll: '{'	 put: #xLbrace;
	atAll: '|'	 put: #xBar;
	atAll: '}'	 put: #xRbrace;
	atAll: '~'	 put: #xTilde;
	atAll: '^'	 put: #xHat;
	atAll: '_'	 put: #xLetter;
	from:  $a to: $z put: #xLetter;
	from:  $A to: $Z put: #xLetter;
	from:  $0 to: $9 put: #xDigit;
	from:   0 to: 32 put: #xSpace;
	at:    $"	 put: #xString;
	at:    $'	 put: #xChar;
	at:    $$	 put: #xOther;
	at:    $@	 put: #xOther;
	at:    $\n	 put: #xEOL.
    (directiveTable := FastIdentityDictionary new)
	at: #if		 put: C99DirectiveIf;
	at: #ifdef	 put: C99DirectiveIfdef;
	at: #ifndef	 put: C99DirectiveIfndef;
	at: #elif	 put: C99DirectiveElif;
	at: #else	 put: C99DirectiveElse;
	at: #endif	 put: C99DirectiveEndif;
	at: #include	 put: C99DirectiveInclude;
	at: #define	 put: C99DirectiveDefine;
	at: #undef	 put: C99DirectiveUndef;
	at: #line	 put: C99DirectiveLine;
	at: #error	 put: C99DirectiveError;
	at: #pragma	 put: C99DirectivePragma.
    stringTable := Array new: 256.
    0 to: 255 do: [:c | stringTable at: c put: (String with: c)].
    text := WriteStream on: (String new: 32).
    bol := true.
    tokenLineNumber := 0.
]

C99Scanner refill
[
    tokenLineNumber == 0 ifTrue: [tokenLineNumber := 1].
    ^super refill
]

" 6.4  Lexical elements

  The categories of preprocessing tokens are: header names, identifiers,
  preprocessing numbers, character constants, string literals, punctuators, and
  single non-white-space characters that do not lexically match the other
  preprocessing token categories.  ... "

C99Scanner identifier			[ bol := false.  ^C99Identifier withText: text contents asSymbol line: tokenLineNumber 		   ]
C99Scanner number			[ bol := false.  ^C99Number	withText: text contents		 line: tokenLineNumber  	   ]
C99Scanner blank			[		 ^C99Blank	withText: text contents		 line: tokenLineNumber  	   ]
C99Scanner blank: aNote			[		 ^C99Blank	withText: text contents		 line: tokenLineNumber note: aNote ]
C99Scanner punctuator: tokText op: type	[ bol := false.  ^type		withText: tokText		 line: tokenLineNumber 		   ]
C99Scanner other: char			[ bol := false.  ^C99Other	withText: (stringTable at: char) line: tokenLineNumber 		   ]

C99Scanner scanText: category until: terminator	"character and string"
[
    | start c |
    bol := false.
    start := line position.
    self next.
    [(c := self next) notNil and: [c ~~ $\n]]
	whileTrue:
	   [c == terminator ifTrue: [^category withText: text contents line: tokenLineNumber].
	    text nextPut: c.
	    (c == $\\ and: [(c := self peek) notNil and: [c ~~ $\n]]) ifTrue: [text nextPut: self next]].
    "...  If a ' or a '' character matches the last category, the behavior is undefined. "
    line position: start.
    ^self other: self next
]

" 5.1.1.2  Translation phases

  3. The source file is decomposed into preprocessing tokens and sequences of 
     white-space characters (including comments).  A source file shall not end in a 
     partial preprocessing token or in a partial comment.  Each comment is replaced by 
     one space character.  New-line characters are retained.  Whether each nonempty 
     sequence of white-space characters other than new-line is retained or replaced by 
     one space character is implementation-defined. "

C99Scanner nextToken
[
    tokenLineNumber := lineNumber.
    (thisChar := self peek) notNil ifTrue: [^self perform: (characterTable at: thisChar)].
    activeGroup notNil ifTrue: [activeGroup error: 'unterminated #if'].
    ^nil
]

" white-space "

C99Scanner xSpace
[
    | c |
    text reset; nextPut: self next.
    [(c := self peek) notNil and: [(characterTable at: c) == #xSpace]] whileTrue: [text nextPut: self next].
    ^self blank
]

C99Scanner xEOL
[
    bol := true.
    text reset; nextPut: self next.
    ^self blank: #eol
]

C99Scanner multiComment		"sent from /"
[
    text reset; nextPut: self next; nextPut: self next.
    [self peek; match: '*/'] whileFalse: [text nextPut: self next].
    text nextPut: self next; nextPut: self next.
    ^self blank: #comment
]

C99Scanner uniComment		"sent from /"
[
    | c |
    text reset; nextPut: self next; nextPut: self next.
    [(c := self peek) notNil and: [c ~~ $\n]] whileTrue: [text nextPut: self next].
    ^self blank: #comment
]

" single non-white-space "

C99Scanner xOther	[ ^self other: self next ]
C99Scanner xIllegal	[ ^self other: self next ]

" character constants and string literals "

C99Scanner xChar	[ text reset.  ^self scanText: C99Character until: $']
C99Scanner xString	[ text reset.  ^self scanText: C99String    until: $"]

" identifiers "

C99Scanner xLetter
[
    | c |
    text reset.
    thisChar == $L
	ifTrue:
	   [line peek: 1 == $" ifTrue: [text nextPut: self next.  ^self scanText: C99String    until: $"].
	    line peek: 1 == $' ifTrue: [text nextPut: self next.  ^self scanText: C99Character until: $']].
    [(c := self peek) notNil and: [(characterTable at: c) == #xLetter or: [(characterTable at: c) == #xDigit]]]
	whileTrue: [text nextPut: self next].
    ^self identifier
]

" pp-number :
	. digit "

C99Scanner xDot
[
    | d |
    text reset.
    ((d := line peek: 1) notNil and: [(characterTable at: d) == #xDigit]) ifTrue: [^self xDigit].
    (self skip: '...') ifTrue: [^self punctuator: '...' op: C99Ellipsis].
    ^self next; punctuator: '.' op: C99Member
]

" pp-number :
	digit "

C99Scanner xDigit
[
    text reset.
    [self ppNumber] whileTrue.
    ^self number.
]

" pp-number :
	pp-number digit
	pp-number identifier-nondigit
	pp-number e sign
	pp-number E sign
	pp-number p sign
	pp-number P sign
	pp-number . "

C99Scanner ppNumber
[
    | c d |
    (c := self peek) isNil ifTrue: [^false].
    (characterTable at: c) == #xDigit  ifTrue: [text nextPut: self next.  ^true].
    (characterTable at: c) == #xLetter ifTrue: [text nextPut: self next.  ^true].
    c == $. ifTrue: [text nextPut: self next.  ^true].
    (d := line peek: 1) isNil ifTrue: [^false].
    ((c == $e or: [c == $E or: [c == $p or: [c == $P]]]) and: [d == $+ or: [d == $-]])
	ifTrue: [text nextPut: self next; nextPut: self next.  ^true].
    ^false
]

" punctuators "

C99Scanner xLparen				[ ^self next; 			     punctuator: '('   op: C99OpenParen ]
C99Scanner xRparen				[ ^self next; 			     punctuator: ')'   op: C99CloseParen ]
C99Scanner xComma				[ ^self next; 			     punctuator: ','   op: C99Comma ]
C99Scanner xSemicolon				[ ^self next; 			     punctuator: ';'   op: C99Semicolon ]
C99Scanner xQuery				[ ^self next; 			     punctuator: '?'   op: C99Query ]
C99Scanner xLbracket				[ ^self next; 			     punctuator: '['   op: C99OpenBracket ]
C99Scanner xRbracket				[ ^self next; 			     punctuator: ']'   op: C99CloseBracket ]
C99Scanner xLbrace				[ ^self next; 			     punctuator: '{'   op: C99Begin ]
C99Scanner xRbrace				[ ^self next; 			     punctuator: '}'   op: C99End ]
C99Scanner xTilde				[ ^self next; 			     punctuator: '~'   op: C99Complement ]
C99Scanner xSlash	" /* // /= / "		[ (self match: '/*' ) ifTrue: [^self multiComment].
						  (self match: '//' ) ifTrue: [^self uniComment].
						  (self skip:  '/=' ) ifTrue: [^self punctuator: '/='  op: C99DivideInto].
						  ^self next;			     punctuator: '/'   op: C99Divide]
C99Scanner xLangle	" <<= << <= <% <: < "	[ (self skip: '<<=' ) ifTrue: [^self punctuator: '<<=' op: C99LeftShiftInto].
						  (self skip: '<<'  ) ifTrue: [^self punctuator: '<<'  op: C99LeftShift].
						  (self skip: '<='  ) ifTrue: [^self punctuator: '<='  op: C99LessEqual].
						  (self skip: '<%'  ) ifTrue: [^self punctuator: '{'   op: C99Begin].
						  (self skip: '<:'  ) ifTrue: [^self punctuator: '['   op: C99OpenBracket].
						  ^self next;			     punctuator: '<'   op: C99Less]
C99Scanner xRangle	" >>= >> >= > "		[ (self skip: '>>=' ) ifTrue: [^self punctuator: '>>=' op: C99RightShiftInto].
						  (self skip: '>>'  ) ifTrue: [^self punctuator: '>>'  op: C99RightShift].
						  (self skip: '>='  ) ifTrue: [^self punctuator: '>='  op: C99GreaterEqual].
						  ^self next;			     punctuator: '>'   op: C99Greater]
C99Scanner xAmpersand	" && &= & "		[ (self skip: '&&'  ) ifTrue: [^self punctuator: '&&'  op: C99LogAnd].
						  (self skip: '&='  ) ifTrue: [^self punctuator: '&='  op: C99AndInto].
						  ^self next;			     punctuator: '&'   op: C99And]
C99Scanner xBar		" || |= | "		[ (self skip: '||'  ) ifTrue: [^self punctuator: '||'  op: C99LogOr].
						  (self skip: '|='  ) ifTrue: [^self punctuator: '|='  op: C99OrInto].
						  ^self next;			     punctuator: '|'   op: C99Or]
C99Scanner xStar	" *= * "		[ (self skip: '*='  ) ifTrue: [^self punctuator: '*='  op: C99MultiplyInto].
						  ^self next;			     punctuator: '*'   op: C99Multiply]
C99Scanner xMinus	" -- -= -> - "		[ (self skip: '--'  ) ifTrue: [^self punctuator: '--'  op: C99Decrement].
						  (self skip: '-='  ) ifTrue: [^self punctuator: '-='  op: C99SubtractInto].
						  (self skip: '->'  ) ifTrue: [^self punctuator: '->'  op: C99MemberRef].
						  ^self next;			     punctuator: '-'   op: C99Subtract]
C99Scanner xPlus	" ++ += + "		[ (self skip: '++'  ) ifTrue: [^self punctuator: '++'  op: C99Increment].
						  (self skip: '+='  ) ifTrue: [^self punctuator: '+='  op: C99AddInto].
						  ^self next;			     punctuator: '+'   op: C99Add]
C99Scanner xPling	" != ! "		[ (self skip: '!='  ) ifTrue: [^self punctuator: '!='  op: C99NotEqual].
						  ^self next;			     punctuator: '!'   op: C99Not]
C99Scanner xColon	" :> : "		[ (self skip: ':>'  ) ifTrue: [^self punctuator: ']'   op: C99CloseBracket].
						  ^self next;			     punctuator: ':'   op: C99Colon]
C99Scanner xEqual	" == = "		[ (self skip: '=='  ) ifTrue: [^self punctuator: '=='  op: C99Equal].
						  ^self next;			     punctuator: '='   op: C99Assign]
C99Scanner xPercent	" %:%: %: %> %= % "	[ (self skip: '%:%:') ifTrue: [^self punctuator: '##'  op: C99Paste].
						  (self skip: '%:'  ) ifTrue: [^self punctuator: '#'   op: C99Hash].
						  (self skip: '%>'  ) ifTrue: [^self punctuator: '}'   op: C99End].
						  (self skip: '%='  ) ifTrue: [^self punctuator: '%='  op: C99ModulusInto].
						  ^self next;			     punctuator: '%'   op: C99Modulus]
C99Scanner xHat		" ^= ^ "		[ (self skip: '^='  ) ifTrue: [^self punctuator: '^='  op: C99XorInto].
						  ^self next;			     punctuator: '^'   op: C99Xor]
C99Scanner xHash	" ## # "		[ (self skip: '##'  ) ifTrue: [^self punctuator: '##'  op: C99Paste].
						   bol		      ifTrue: [^self directive             ].
						  ^self next;			     punctuator: '#'   op: C99Hash]

" directives "

C99Scanner directive
[
    | start token dir |
    bol ifFalse: [self error: 'directive processing reentered after punt'].
    bol := false.
    start := line position.
    self next.
    [(token := self nextToken) isBlank and: [token notEOL]] whileTrue.
    (token isIdentifier and: [(dir := directiveTable at: token text ifAbsent: []) notNil])
	ifFalse:
	   [line position: start.	"tokenize again ignoring directives"
	    ^self nextToken].
    (dir := dir withText: token text line: tokenLineNumber)
	name: token;
	scanFrom: self;
	doDirectiveIn: self.
    ^dir
]

C99Directive scanFrom: scanner
[
    | tok |
    [(tok := scanner nextToken) notNil and: [tok notEOL]] whileTrue: [body add: tok]
]

C99Scanner doDirectiveIf: token
[
    token enclosingGroup: activeGroup.
    activeGroup := token.
]

C99Scanner doDirectiveIfdef: token
[
    token enclosingGroup: activeGroup.
    activeGroup := token.
]

C99Scanner doDirectiveIfndef: token
[
    token enclosingGroup: activeGroup.
    activeGroup := token.
]

C99Scanner doDirectiveElif: token
[
    activeGroup isNil ifTrue: [token error: '#elif without #if'].
    activeGroup isDirectiveElse ifTrue: [token error: '#elif follows #else'].
    activeGroup nextGroup: token.
    token enclosingGroup: activeGroup enclosingGroup.
    activeGroup := token.
]

C99Scanner doDirectiveElse: token
[
    activeGroup isNil ifTrue: [token error: '#else without #if'].
    activeGroup isDirectiveElse ifTrue: [token error: '#else follows #else'].
    activeGroup nextGroup: token.
    token enclosingGroup: activeGroup enclosingGroup.
    activeGroup := token.
]

C99Scanner doDirectiveEndif: token
[
    activeGroup isNil ifTrue: [token error: '#endif without #if'].
    activeGroup nextGroup: token.
    activeGroup := activeGroup enclosingGroup.
]

C99Scanner doDirectiveInclude: token []
C99Scanner doDirectiveDefine:  token []
C99Scanner doDirectiveUndef:   token []
C99Scanner doDirectiveLine:    token []
C99Scanner doDirectiveError:   token []
C99Scanner doDirectivePragma:  token []

"----------------------------------------------------------------"

C99Preprocessor : C99TokenList ( macros currentToken )

C99Preprocessor new
[
    self := super new.
    macros := FastIdentityDictionary new.
]

" 5.1.1.2  Translation phases

  4. Preprocessing directives are executed, macro invocations are expanded, and 
     _Pragma unary operator expressions are executed.  If a character sequence that 
     matches the syntax of a universal character name is produced by token 
     concatenation (6.10.3.3), the behavior is undefined.  A #include preprocessing 
     directive causes the named header or source file to be processed from phase 1 
     through phase 4, recursively.  All preprocessing directives are then deleted. "

C99Preprocessor nextToken
[
   [self isEmpty ifTrue: [^nil].
    (currentToken := self removeFirst) doDirectiveOrIdentifierIn: self.
    currentToken isNil] whileTrue.	"token deleted itself"
   ^currentToken
]

C99Token	doDirectiveOrIdentifierIn: anObject	[ ]
C99Directive	doDirectiveOrIdentifierIn: anObject	[ ^self doDirectiveIn: anObject ]
C99Identifier	doDirectiveOrIdentifierIn: anObject	[ ^anObject doIdentifier: self ]

"
C99Preprocessor doDirectivePragma:  token []
C99Preprocessor doDirectiveLine:    token []
C99Preprocessor doDirectiveError:   token []
C99Preprocessor doDirectiveInclude: token []
"

" conditional sections "

C99Preprocessor doDirectiveIf: token		[ self ifSection: token ]
C99Preprocessor doDirectiveIfdef: token		[ self ifSection: token ]
C99Preprocessor doDirectiveIfndef: token	[ self ifSection: token ]

C99Preprocessor ifSection: token
[
    [token isDirectiveEndif or: [token evaluateConditionIn: self]] whileFalse: [token := self skipTo: token nextGroup].
    currentToken := nil.	"discard the directive (first active #if, #elif, #else or final #endif)"
]

C99Preprocessor doDirectiveElif: token
[
    [token isDirectiveEndif] whileFalse: [token := self skipTo: token nextGroup].
    currentToken := nil.	"discard #endif"
]

C99Preprocessor doDirectiveEndif: token
[
    currentToken := nil.	"discard"
]

C99Preprocessor skipTo: token
[
    [currentToken == token or: [self isEmpty]] whileFalse: [currentToken := self removeFirst].
    self isEmpty ifTrue: [token error: 'missing #endif'].
    ^currentToken		"token will be returned to client unless discarded"
]

C99DirectiveIf evaluateConditionIn: cpp		"inherited by C99DirectiveElif"
[
    | id |
    self removeBlanks.
    body isEmpty ifTrue: [self error: '#if with no expression'].
    ^(body expandMacrosIn: cpp) evaluate ~~ 0
]

C99DirectiveIfdef evaluateConditionIn: cpp	"negated by C99DirectiveIfndef"
[
    | id |
    self removeBlanks.
    (body notEmpty and: [body first isIdentifier])	ifFalse: [self error: '#', text, ' requires an identifier'].
    (body size == 1)					ifFalse: [self warn: 'ignoring junk after #', text].
    ^(cpp lookup: id) notNil
]

C99Preprocessor lookup: name	[ ^macros at: name ifAbsent: [] ]

C99DirectiveIfndef evaluateConditionIn: cpp
[
    ^(super evaluateConditionIn: cpp) not
]

C99DirectiveElse evaluateConditionIn: cpp
[
    ^true
]

" macro un/definition "

C99Preprocessor doDirectiveUndef: token
[
    | args id |
    args := token body removeOuterBlanks combineBlanks.
    (args notEmpty and: [args first isIdentifier]) ifFalse: [token error: '#undef requires an identifier'].
    id := args removeFirst.
    macros removeKey: id ifNotFound: []
]

C99Preprocessor doDirectiveDefine: token
[
    | macro defn |
    macro := self createMacro: token.
    defn := macros at: macro name ifAbsent: [macros at: macro name put: macro].
    macro = defn ifFalse: [token error: 'macro ', macro name, ' redefined'].
    currentToken := nil.
]

C99Preprocessor createMacro: token
[
    | args id |
    args := token body removeOuterBlanks combineBlanks.
    (args notEmpty and: [args first isIdentifier]) ifFalse: [token error: '#define requires an identifier'].
    id := args removeFirst.
    ^((args notEmpty and: [args first isOpenParen]) ifTrue: [C99FunctionMacro] ifFalse: [C99ObjectMacro])
	withName: id text
	token:    token
	args:	  args
]

C99Macro : Object ( name token replacementList )
  C99ObjectMacro : C99Macro ()
  C99FunctionMacro : C99Macro ( argumentList )

C99Macro name			[ ^name ]
C99Macro replacementList	[ ^replacementList ]

C99Macro withName: nameSymbol token: defineToken replacementList: tokenList
[
    self := self new.
    name := nameSymbol.
    token := defineToken.
    replacementList := tokenList.
]

C99ObjectMacro withName: nameSymbol token: defineToken args: args
[
    ^self withName: nameSymbol token: defineToken replacementList: defineToken body removeOuterBlanks
]

C99ObjectMacro = aMacro
[
    ^self == aMacro or: [aMacro isC99ObjectMacro and: [replacementList = aMacro replacementList]]
]

C99FunctionMacro withName: nameSymbol token: defineToken args: args
[
    self := self withName: nameSymbol token: defineToken replacementList: defineToken body removeOuterBlanks
]

C99FunctionMacro = aMacro
[
    ^self == aMacro or:	[aMacro isC99FunctionMacro
			     and: [replacementList = aMacro replacementList
			     and: [argumentList = aMacro argumentList]]]
]

" macro expansion "

C99Preprocessor doIdentifier: token
[
    | macro |
    (macro := macros at: token text ifAbsent: []) isNil ifTrue: [^self].
    macro expand: token before: self in: self.
    currentToken := nil.
]

"----------------------------------------------------------------"

C99Keyword : C99Identifier ()

C99Auto : C99Keyword ()
C99Break : C99Keyword ()
C99Case : C99Keyword ()
C99Char : C99Keyword ()
C99Const : C99Keyword ()
C99Continue : C99Keyword ()
C99Default : C99Keyword ()
C99Do : C99Keyword ()
C99Double : C99Keyword ()
C99Else : C99Keyword ()
C99Enum : C99Keyword ()
C99Extern : C99Keyword ()
C99Float : C99Keyword ()
C99For : C99Keyword ()
C99Goto : C99Keyword ()
C99If : C99Keyword ()
C99Inline : C99Keyword ()
C99Int : C99Keyword ()
C99Long : C99Keyword ()
C99Register : C99Keyword ()
C99Restrict : C99Keyword ()
C99Return : C99Keyword ()
C99Short : C99Keyword ()
C99Signed : C99Keyword ()
C99Sizeof : C99Keyword ()
C99Static : C99Keyword ()
C99Struct : C99Keyword ()
C99Switch : C99Keyword ()
C99Typedef : C99Keyword ()
C99Union : C99Keyword ()
C99Unsigned : C99Keyword ()
C99Void : C99Keyword ()
C99Volatile : C99Keyword ()
C99While : C99Keyword ()
C99Bool : C99Keyword ()
C99Complex : C99Keyword ()
C99Imaginary : C99Keyword ()


C99Parser : C99TokenList ( tokens )

C99Parser on: tokenList
[
    self := super new.
    tokens := tokenList.
]

" 5.1.1.2 Translation phases

  5. Each source character set member and escape sequence in character constants and 
     string literals is converted to the corresponding member of the execution character 
     set; if there is no corresponding member, it is converted to an implementation- 
     defined member other than the null (wide) character. 

  6. Adjacent string literal tokens are concatenated. 

  7. White-space characters separating tokens are no longer significant. Each 
     preprocessing token is converted into a token. The resulting tokens are 
     syntactically and semantically analyzed and translated as a translation unit. "

C99Parser nextToken
[
    | token |
    [ tokens isEmpty ifTrue: [^nil].
      (token := tokens removeFirst createTokenBefore: tokens) isNil ] whileTrue.
    ^token
]

C99Token createTokenBefore: tokenList	[ ^self ]

C99Blank createTokenBefore: tokenList	[ ^nil ]	"discard white-space"

C99String createTokenBefore: tokenList			"concatenate strings"
[
    [tokenList notEmpty and: [tokenList first isC99String]]
	whileTrue:
	   [self combineWith: tokenList removeFirst].
    ^self.
] 

C99Directive createTokenBefore: tokenList		"discard directives"
[
    self warn: 'this token should not have escaped the preprocessor'.
    ^nil
]

C99Number createTokenBefore: tokenList			"convert pp-number to proper token"
[
    ^self
]

"----------------------------------------------------------------"

[
    | v1 v2 v3 v4 line scanner token cpp |
    v1 := v2 := v3 := v4 := false.
    OS arguments do: [:arg |
	arg = '-1' ifTrue: [v1 := true].
	arg = '-2' ifTrue: [v2 := true].
	arg = '-3' ifTrue: [v3 := true].
	arg = '-4' ifTrue: [v4 := true]].
    scanner := C99Scanner on: StdIn.
    cpp := C99Preprocessor new.
    [token := scanner nextToken]
	whileTrue:
	   [v1 ifTrue: [StdOut print: token; cr].
	    cpp add: token].
    v2 ifTrue: [cpp do: [:t | t putTextOn: StdOut]].
    line := 1.
    [token := cpp nextToken]
	whileTrue:
	   [v3 ifTrue: [StdOut print: token; cr].
	    v4 ifTrue:
	       [token isEOL ifTrue: [line := line + 1].
		[line < token line] whileTrue: [line := line + 1.  StdOut cr].
		token putTextOn: StdOut]].
]
