{ import: Window-lib }

Matrix4x4 isMatrix4x4	[ ^true ]
Object isMatrix4x4	[ ^false ]

Matrix4x4 _sizeof	{ return (oop)(sizeof(double) * 16); }

Matrix4x4
    with: xx with: xy with: xz with: xw
    with: yx with: yy with: yz with: yw
    with: zx with: zy with: zz with: zw
    with: wx with: wy with: wz with: ww
[
    (self := self new)
	at: 0 at: 0 put: xx;  at: 0 at: 1 put: xy;  at: 0 at: 2 put: xz;  at: 0 at: 3 put: xw;
	at: 1 at: 0 put: yx;  at: 1 at: 1 put: yy;  at: 1 at: 2 put: yz;  at: 1 at: 3 put: yw;
	at: 2 at: 0 put: zx;  at: 2 at: 1 put: zy;  at: 2 at: 2 put: zz;  at: 2 at: 3 put: zw;
	at: 3 at: 0 put: wx;  at: 3 at: 1 put: wy;  at: 3 at: 2 put: wz;  at: 3 at: 3 put: ww.
]

Matrix4x4 at: i at: j put: aFloat
[
    {
        int i= (long)v_i >> 1, j= (long)v_j >> 1;
        if (((long)v_i & (long)v_j & 1) && (i >= 0) && (i <= 3) && (j >= 0) && (j <= 3)) {
	    memcpy(&((double *)v_self)[i * 4 + j], (double *)v_aFloat, sizeof(double));
	    return v_aFloat;
        }
    }.
    ^super at: i at: j put: aFloat
]

Matrix4x4 at: i at: j
[
    | result |
    result := Float new.
    {
        int i= (long)v_i >> 1, j= (long)v_j >> 1;
        if (((long)v_i & (long)v_j & 1) && (i >= 0) && (i <= 3) && (j >= 0) && (j <= 3)) {
	    memcpy((double *)v_result, &((double *)v_self)[i * 4 + j], sizeof(double));
            return v_result;
        }
    }.
    ^super at: i at: j
]

Matrix4x4 newZero
[
    self := super new.
    0 to: 3 do: [:i | 0 to: 3 do: [:j | self at: i at: j put: 0.0]].
]

Matrix4x4Zero := [ Matrix4x4 newZero ]

Matrix4x4 zero	[ ^Matrix4x4Zero ]

Matrix4x4 newIdentity
[
    self := super new.
    0 to: 3 do: [:i | 0 to: 3 do: [:j | self at: i at: j put: 0.0]].
    0 to: 3 do: [:i | self at: i at: i put: 1.0].
]

Matrix4x4Identity := [ Matrix4x4 newIdentity ]

Matrix4x4 identity	[ ^Matrix4x4Identity ]

Matrix4x4 * aMatrixOrVector
[
    aMatrixOrVector isMatrix4x4 ifTrue: [^self multipliedByMatrix4x4: aMatrixOrVector ].
    aMatrixOrVector isVector4   ifTrue: [^self multipliedByVector4:   aMatrixOrVector ].
    ^super * aMatrixOrVector
]

Matrix4x4 multipliedByMatrix4x4: aMatrix
[
    | product |
    product := self new.
    0 to: 3 do: [:i | 0 to: 3 do: [:j |
	product at: i at: j put:
	    ((self at: i at: 0) * (aMatrix at: 0 at: j)) +
	    ((self at: i at: 1) * (aMatrix at: 1 at: j)) +
	    ((self at: i at: 2) * (aMatrix at: 2 at: j)) +
	    ((self at: i at: 3) * (aMatrix at: 3 at: j))]].
    ^product.
]

Matrix4x4 multipliedByVector4: aVector
[
    | product |
    product := aVector new.
    0 to: 3 do: [:i |
	product at: i put:
	    ((self at: i at: 0) * (aVector at: 0)) +
	    ((self at: i at: 1) * (aVector at: 1)) +
	    ((self at: i at: 2) * (aVector at: 2)) +
	    ((self at: i at: 3) * (aVector at: 3))].
    ^product.
]

Matrix4x4 printOn: aStream
[
    aStream nextPutAll: 'Matrix4x4('.
    0 to: 3 do: [:i | aStream cr.  0 to: 3 do: [:j | aStream space; print: (self at: i at: j)]].
    aStream nextPutAll: ' )'.
]

"----------------------------------------------------------------"

Vector4 isVector4	[ ^true ]
Object isVector4	[ ^false ]

Vector4 _sizeof	{ return (oop)(sizeof(double) * 4); }

Vector4 x	[ ^self at: 0 ]
Vector4 y	[ ^self at: 1 ]
Vector4 z	[ ^self at: 2 ]
Vector4 w	[ ^self at: 3 ]

Vector4 at: i put: aFloat
[
    {
        int i= (long)v_i >> 1;
        if (((long)v_i & 1) && (i >= 0) && (i <= 3)) {
	    memcpy(&((double *)v_self)[i], (double *)v_aFloat, sizeof(double));
	    return v_aFloat;
        }
    }.
    ^super at: i put: aFloat
]

Vector4 at: i
[
    | result |
    result := Float new.
    {
        int i= (long)v_i >> 1;
        if (((long)v_i & 1) && (i >= 0) && (i <= 3)) {
	    memcpy((double *)v_result, &((double *)v_self)[i], sizeof(double));
            return v_result;
        }
    }.
    ^super at: i
]

Vector4 newZero
[
    self := super new.
    0 to: 3 do: [:i | self at: i put: 0.0].
]

Vector4Zero := [ Vector4 newZero ]

Vector4 zero	[ ^Vector4Zero ]

Vector4 with: x with: y with: z with: w
[
    (self := self new)
	at: 0 put: x;
	at: 1 put: y;
	at: 2 put: z;
	at: 3 put: w.
]

Vector4 + aVector
[
    ^self
	with: (self at: 0) + (aVector at: 0)
	with: (self at: 1) + (aVector at: 1)
	with: (self at: 2) + (aVector at: 2)
	with: (self at: 3) + (aVector at: 3)
]

Vector4 printOn: aStream
[
    aStream nextPutAll: 'Vector4('.
    0 to: 3 do: [:i | aStream space; print: (self at: i)].
    aStream nextPutAll: ' )'.
]
