{ import: Xlib }

Event : Object ( xEvent position handled )

MotionEvent : Event ()
ExposeEvent : Event ()
ResizeEvent : Event ( width height )
InputEvent : Event ( state )
  KeyEvent : InputEvent ( keyCode )
    KeyPressEvent : KeyEvent ()
    KeyReleaseEvent : KeyEvent ()
  ButtonEvent : InputEvent ( button )
    ButtonPressEvent : KeyEvent ()
    ButtonReleaseEvent : KeyEvent ()

EventTypeMap := [ IdentityDictionary new
		      at: Xlib MotionNotify	put: MotionEvent;
		      at: Xlib ButtonPress	put: ButtonPressEvent;
		      at: Xlib ButtonRelease	put: ButtonReleaseEvent;
		      at: Xlib KeyPress		put: KeyPressEvent;
		      at: Xlib KeyRelease	put: KeyReleaseEvent;
		      at: Xlib Expose		put: ExposeEvent;
		      at: Xlib ConfigureNotify	put: ResizeEvent;
		      yourself ]

NoEvent : Object ()

NoEvent newWithXEvent: aXEvent	[ ^nil ]

Event withXEvent: aXEvent
[
    ^(EventTypeMap at: aXEvent type ifAbsent: [NoEvent]) newWithXEvent: aXEvent
]

Event newWithXEvent: aXEvent
[
    self := super new.
    xEvent := aXEvent.
    position := xEvent position.
    handled := false
]

ResizeEvent width	[ ^width ]
ResizeEvent height	[ ^height ]

ResizeEvent newWithXEvent: aXEvent
[
    self := super newWithXEvent: aXEvent.
    width  := aXEvent width.
    height := aXEvent height.
]

InputEvent state	[ ^state ]

InputEvent newWithXEvent: aXEvent
[
    self := super newWithXEvent: aXEvent.
    state   := aXEvent state.
]

KeyEvent keyCode	[ ^keyCode ]

KeyEvent newWithXEvent: aXEvent
[
    self := super newWithXEvent: aXEvent.
    keyCode := aXEvent keyCode.
]

ButtonEvent button	[ ^button ]

ButtonEvent newWithXEvent: aXEvent
[
    self := super newWithXEvent: aXEvent.
    button := aXEvent button.
]

Event position		[ ^position ]
Event type		[ ^nil ]

Event isMotion		[ ^false ]	MotionEvent		isMotion	[ ^true ]	MotionEvent		type [ ^#Motion		]
Event isButtonPress	[ ^false ]	ButtonPressEvent	isButtonPress	[ ^true ]	ButtonPressEvent	type [ ^#ButtonPress	]
Event isButtonRelease	[ ^false ]	ButtonReleaseEvent	isButtonRelease	[ ^true ]	ButtonReleaseEvent	type [ ^#ButtonRelease  ]
Event isKeyPress	[ ^false ]	KeyPressEvent		isKeyPress	[ ^true ]	KeyPressEvent		type [ ^#KeyPress	]
Event isKeyRelease	[ ^false ]	KeyReleaseEvent		isKeyRelease	[ ^true ]	KeyReleaseEvent		type [ ^#KeyRelease	]
Event isExpose		[ ^false ]	ExposeEvent		isExpose	[ ^true ]	ExposeEvent		type [ ^#Expose  	]
Event isResize		[ ^false ]	ResizeEvent		isResize	[ ^true ]	ResizeEvent		type [ ^#Resize  	]

"----------------------------------------------------------------"

Surface : Object ( context )

Surface withContext: aContext
[
    self := self new.
    context := aContext.
]

Surface new: boundingRectangle
[
    self := self new.
    context := (X11Display open rootWindow subWindow: boundingRectangle) map newGC.
    [self waitEvent isExpose] whileFalse.
]

Surface newPixmap: boundingRectangle
[
    ^self withContext: (context drawable newPixmap: boundingRectangle) newGC.
]

Surface context	[ ^context ]

Surface close
[
    context drawable unmap; destroy.
    context free.
]

Surface width					[ ^context drawable width ]
Surface height					[ ^context drawable height ]
Surface depth					[ ^context drawable depth ]
Surface extent					[ ^self width @ self height ]

Surface foreground: aPixel			[ context foreground: aPixel ]
Surface drawPoint: aPoint			[ context drawPoint: aPoint ]
Surface drawLine: startPoint to: endPoint	[ context drawLine: startPoint to: endPoint ]
Surface drawRectangle: aRectangle		[ context drawRectangle: aRectangle ]
Surface fillRectangle: aRectangle		[ context fillRectangle: aRectangle ]
Surface clear: aPixel				[ context drawable background: aPixel.  self clear ]
Surface clear					[ self clearArea: (0@0 corner: self extent) ]
Surface clearArea: aRectangle			[ context drawable clearArea: aRectangle ]
Surface sync					[ context drawable display sync ]
Surface flush					[ context drawable display flush ]

Surface drawArc: centerPoint extent: extentPoint from: startAngle to: endAngle
[
    context drawArc: centerPoint extent: extentPoint from: startAngle to: endAngle
]

Surface fillArc: centerPoint extent: extentPoint from: startAngle to: endAngle
[
    context fillArc: centerPoint extent: extentPoint from: startAngle to: endAngle
]

Surface drawOn: aSurface at: aPoint
[
    context copyArea: context drawable to: aSurface context drawable from: 0@0 extent: self extent to: aPoint
]

Surface drawString: aString at: aPoint
[
    context drawString: aString at: aPoint
]

SurfaceStringBuffer := [ String new: 1 ]

Surface drawCharacter: aCharacter at: aPoint
[
    SurfaceStringBuffer at: 0 put: aCharacter.
    self drawString: SurfaceStringBuffer at: aPoint
]

Surface lineWidth: newWidth
[
    | xgcv |
    xgcv := XGCValues new.
    xgcv lineWidth: newWidth.
    context changeGC: Xlib GCLineWidth values: xgcv.
]

Surface font: newFont
[
    | xgcv |
    xgcv := XGCValues new.
    xgcv font: newFont _id.
    context changeGC: Xlib GCFont values: xgcv.
]

Surface pollEvent
[
    | xEvent |
    ^(xEvent := context drawable checkEvent) notNil ifTrue: [Event withXEvent: xEvent]
]

Surface waitEvent
[
    | event |
    [(event := Event withXEvent: context drawable nextEvent) isNil] whileTrue.
    ^event
]

Surface fontNamed: nameString			[ ^context drawable display fontNamed: nameString ]

Surface defaultFont				[ ^self fontNamed: 'fixed' ]

Surface beDoubleBuffered
[
    context drawable beDoubleBuffered.
    context := context drawable newGC.
    
]

Surface beginDrawing				[ context drawable beginDrawing ]
Surface endDrawing				[ context drawable swapBuffers; endDrawing ]
