;;; coke-mode.el --- Coke editing mode
;;; 
;;; based shamelessly on scheme.el

(require 'lisp-mode)

(defvar coke-mode-syntax-table
  (let ((st (make-syntax-table))
	(i 0))

    (while (< i 256)
      (modify-syntax-entry i "_   " st)
      (setq i (1+ i)))

    (setq i ?0)
    (while (<= i ?9)
      (modify-syntax-entry i "w   " st)
      (setq i (1+ i)))
    (setq i ?A)
    (while (<= i ?Z)
      (modify-syntax-entry i "w   " st)
      (setq i (1+ i)))
    (setq i ?a)
    (while (<= i ?z)
      (modify-syntax-entry i "w   " st)
      (setq i (1+ i)))

    (modify-syntax-entry ?\t "    " st)
    (modify-syntax-entry ?\n ">   " st)
    (modify-syntax-entry ?\f "    " st)
    (modify-syntax-entry ?\r "    " st)
    (modify-syntax-entry ?\s "    " st)
    (modify-syntax-entry ?\[ "(]  " st)
    (modify-syntax-entry ?\] ")[  " st)
    (modify-syntax-entry ?{ "(}  " st)
    (modify-syntax-entry ?} "){  " st)
    (modify-syntax-entry ?\| "\" 23bn" st)
    (modify-syntax-entry ?\( "()  " st)
    (modify-syntax-entry ?\) ")(  " st)
    (modify-syntax-entry ?\; "< 2 " st)
    (modify-syntax-entry ?\" "\"   " st)
    (modify-syntax-entry ?' "'   " st)
    (modify-syntax-entry ?` "'   " st)
    (modify-syntax-entry ?, "'   " st)
    (modify-syntax-entry ?@ "'   " st)
    (modify-syntax-entry ?# "' 14b" st)
    (modify-syntax-entry ?\\ "\\   " st)
    st))

(defvar coke-mode-abbrev-table nil)
(define-abbrev-table 'coke-mode-abbrev-table ())

(defvar coke-imenu-generic-expression
      '((nil
	 "^(define\\(\\|-\\(generic\\(\\|-procedure\\)\\|method\\)\\)*\\s-+(?\\(\\sw+\\)" 4)
	("Types"
	 "^(define-class\\s-+(?\\(\\sw+\\)" 1)
	("Macros"
	 "^(\\(defmacro\\|define-macro\\|define-syntax\\)\\s-+(?\\(\\sw+\\)" 2))
  "Imenu generic expression for Coke mode.  See `imenu-generic-expression'.")

(defun coke-mode-variables ()
  (set-syntax-table coke-mode-syntax-table)
  (setq local-abbrev-table coke-mode-abbrev-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'fill-paragraph-function)
  (setq fill-paragraph-function 'lisp-fill-paragraph)
  (make-local-variable 'adaptive-fill-mode)
  (setq adaptive-fill-mode nil)
  (make-local-variable 'normal-auto-fill-function)
  (setq normal-auto-fill-function 'lisp-mode-auto-fill)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'lisp-indent-line)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'outline-regexp)
  (setq outline-regexp ";;; \\|(....")
  (make-local-variable 'comment-start)
  (setq comment-start ";")
  (set (make-local-variable 'comment-add) 1)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "\\(\\(^\\|[^\\\\\n]\\)\\(\\\\\\\\\\)*\\);+[ \t]*")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'lisp-indent-function)
  (setq lisp-indent-function 'coke-indent-function)
  (setq mode-line-process '("" coke-mode-line-process))
  (set (make-local-variable 'imenu-case-fold-search) t)
  (setq imenu-generic-expression coke-imenu-generic-expression)
  (set (make-local-variable 'imenu-syntax-alist)
	'(("+-*/.<>=?!$%_&~^:" . "w")))
  (set (make-local-variable 'font-lock-defaults)
       '((coke-font-lock-keywords
          coke-font-lock-keywords-1 coke-font-lock-keywords-2)
         nil t (("+-*/.<>=!?$%_&~^:" . "w") (?#. "w 14"))
         beginning-of-defun
         (font-lock-mark-block-function . mark-defun)
         (font-lock-syntactic-face-function
          . coke-font-lock-syntactic-face-function)
         (parse-sexp-lookup-properties . t)
         (font-lock-extra-managed-props syntax-table)))
  (set (make-local-variable 'lisp-doc-string-elt-property)
       'coke-doc-string-elt))

(defvar coke-mode-line-process "")

(defvar coke-mode-map
  (let ((smap (make-sparse-keymap))
	(map (make-sparse-keymap "Coke")))
    (set-keymap-parent smap lisp-mode-shared-map)
    (define-key smap [menu-bar coke] (cons "Coke" map))
    (define-key map [run-coke] '("Run Inferior Coke" . run-coke))
    (define-key map [uncomment-region]
      '("Uncomment Out Region" . (lambda (beg end)
                                   (interactive "r")
                                   (comment-region beg end '(4)))))
    (define-key map [comment-region] '("Comment Out Region" . comment-region))
    (define-key map [indent-region] '("Indent Region" . indent-region))
    (define-key map [indent-line] '("Indent Line" . lisp-indent-line))
    (put 'comment-region 'menu-enable 'mark-active)
    (put 'uncomment-region 'menu-enable 'mark-active)
    (put 'indent-region 'menu-enable 'mark-active)
    smap)
  "Keymap for Coke mode.
All commands in `lisp-mode-shared-map' are inherited by this map.")

(defun coke-mode-commands (map)
  (define-key map "\177" 'backward-delete-char-untabify)
  (define-key map "\e\C-q" 'indent-sexp))

;;;###autoload
(defun coke-mode ()
  "Major mode for editing Coke code.
Editing commands are similar to those of `lisp-mode'.

In addition, if an inferior Coke process is running, some additional
commands will be defined, for evaluating expressions and controlling
the interpreter, and the state of the process will be displayed in the
modeline of all Coke buffers.  The names of commands that interact
with the Coke process start with \"xcoke-\" if you use the MIT
Coke-specific `xcoke' package; for more information see the
documentation for `xcoke-interaction-mode'.  Use \\[run-coke] to
start an inferior Coke using the more general `cmucoke' package.

Commands:
Delete converts tabs to spaces as it moves back.
Blank lines separate paragraphs.  Semicolons start comments.
\\{coke-mode-map}
Entry to this mode calls the value of `coke-mode-hook'
if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map coke-mode-map)
  (setq major-mode 'coke-mode)
  (setq mode-name "Coke")
  (coke-mode-variables)
  (run-mode-hooks 'coke-mode-hook))

(defgroup coke nil
  "Editing Coke code."
  :link '(custom-group-link :tag "Font Lock Faces group" font-lock-faces)
  :group 'lisp)

(defcustom coke-mit-dialect t
  "If non-nil, coke mode is specialized for MIT Coke.
Set this to nil if you normally use another dialect."
  :type 'boolean
  :group 'coke)

(defcustom coke-mode-hook nil
  "Normal hook run when entering `coke-mode'.
See `run-hooks'."
  :type 'hook
  :group 'coke)

(defcustom coke-program-name "coke"
  "*Program invoked by the `run-coke' command."
  :type 'string
  :group 'coke)

(defconst coke-font-lock-keywords-1
  (eval-when-compile
    (list
     (list (concat "(\\(\\(define\\)\\*?\\("
		   "\\(\\|-public\\|-method\\|-generic\\(-procedure\\)?\\)\\|"
		   "\\(-syntax\\|-macro\\)\\|"
		   "-class"
                   "\\|-module"
		   "\\)\\)\\>"
		   "[ \t]*(?"
		   "\\(\\sw+\\)?")
	   '(1 font-lock-keyword-face)
	   '(6 (cond ((match-beginning 3) font-lock-function-name-face)
		     ((match-beginning 5) font-lock-variable-name-face)
		     (t font-lock-type-face))
	       nil t))
     ))
  "Subdued expressions to highlight in Coke modes.")

(defconst coke-font-lock-keywords-2
  (append coke-font-lock-keywords-1
   (eval-when-compile
     (list
      (cons
       (concat
	"(" (regexp-opt
	     '("begin" "call-with-current-continuation" "call/cc"
	       "call-with-input-file" "call-with-output-file" "case" "cond"
	       "do" "else" "for-each" "if" "lambda" "while" "syntax"
	       "let" "let*" "let-syntax" "letrec" "letrec-syntax" "set"
	       "and" "or" "delay" "force"
	       "quasiquote" "quote" "unquote" "unquote-splicing"
	       "map" "syntax" "syntax-rules") t)
	"\\>") 1)
      '("(let\\s-+\\(\\sw+\\)"
        (1 font-lock-function-name-face))
      '("\\<<\\sw+>\\>" . font-lock-type-face)
      '("\\<#?:\\sw+\\>" . font-lock-builtin-face)
      )))
  "Gaudy expressions to highlight in Coke modes.")

(defvar coke-font-lock-keywords coke-font-lock-keywords-1
  "Default expressions to highlight in Coke modes.")

(defconst coke-sexp-comment-syntax-table
  (let ((st (make-syntax-table coke-mode-syntax-table)))
    (modify-syntax-entry ?\; "." st)
    (modify-syntax-entry ?\n " " st)
    (modify-syntax-entry ?#  "'" st)
    st))

(put 'lambda 'coke-doc-string-elt 2)
(put 'define 'coke-doc-string-elt
     (lambda ()
       (forward-comment (point-max))
       (if (eq (char-after) ?\() 2 0)))
(put 'syntax 'coke-doc-string-elt
     (lambda ()
       (forward-comment (point-max))
       (if (eq (char-after) ?\() 2 0)))

(defun coke-font-lock-syntactic-face-function (state)
  (when (and (null (nth 3 state))
             (eq (char-after (nth 8 state)) ?#)
             (eq (char-after (1+ (nth 8 state))) ?\;))
    (save-excursion
      (let ((pos (point))
            (end
             (condition-case err
                 (let ((parse-sexp-lookup-properties nil))
                   (goto-char (+ 2 (nth 8 state)))
                   (forward-sexp 1)
                   (point))
               (scan-error (nth 2 err)))))
        (when (< pos (- end 2))
          (put-text-property pos (- end 2)
                             'syntax-table coke-sexp-comment-syntax-table))
        (put-text-property (- end 1) end 'syntax-table '(12)))))
  (lisp-font-lock-syntactic-face-function state))


(defvar calculate-lisp-indent-last-sexp)

(defun coke-indent-function (indent-point state)
  (let ((normal-indent (current-column)))
    (goto-char (1+ (elt state 1)))
    (parse-partial-sexp (point) calculate-lisp-indent-last-sexp 0 t)
    (if (and (elt state 2)
             (not (looking-at "\\sw\\|\\s_")))
        (progn
          (if (not (> (save-excursion (forward-line 1) (point))
                      calculate-lisp-indent-last-sexp))
              (progn (goto-char calculate-lisp-indent-last-sexp)
                     (beginning-of-line)
                     (parse-partial-sexp (point)
					 calculate-lisp-indent-last-sexp 0 t)))
          (backward-prefix-chars)
          (current-column))
      (let ((function (buffer-substring (point)
					(progn (forward-sexp 1) (point))))
	    method)
	(setq method (or (get (intern-soft function) 'coke-indent-function)
			 (get (intern-soft function) 'coke-indent-hook)))
	(cond ((or (eq method 'defun)
		   (and (null method)
			(> (length function) 3)
			(string-match "\\`def" function)))
	       (lisp-indent-defform state indent-point))
	      ((integerp method)
	       (lisp-indent-specform method state
				     indent-point normal-indent))
	      (method
		(funcall method state indent-point normal-indent)))))))


(defun would-be-symbol (string)
  (not (string-equal (substring string 0 1) "(")))

(defun next-sexp-as-string ()
  (forward-sexp 1)
  (let ((the-end (point)))
    (backward-sexp 1)
    (buffer-substring (point) the-end)))

(defun coke-let-indent (state indent-point normal-indent)
  (skip-chars-forward " \t")
  (if (looking-at "[-a-zA-Z0-9+*/?!@$%^&_:~]")
      (lisp-indent-specform 2 state indent-point normal-indent)
    (lisp-indent-specform 1 state indent-point normal-indent)))

(put 'begin 'coke-indent-function 0)
(put 'case 'coke-indent-function 1)
(put 'delay 'coke-indent-function 0)
(put 'do 'coke-indent-function 2)
(put 'lambda 'coke-indent-function 1)
(put 'while 'coke-indent-function 1)
(put 'syntax 'coke-indent-function 1)
(put 'let 'coke-indent-function 'coke-let-indent)
(put 'let* 'coke-indent-function 1)
(put 'letrec 'coke-indent-function 1)
(put 'sequence 'coke-indent-function 0)
(put 'let-syntax 'coke-indent-function 1)
(put 'letrec-syntax 'coke-indent-function 1)
(put 'syntax-rules 'coke-indent-function 1)
(put 'syntax-case 'coke-indent-function 2)

(put 'call-with-input-file 'coke-indent-function 1)
(put 'with-input-from-file 'coke-indent-function 1)
(put 'with-input-from-port 'coke-indent-function 1)
(put 'call-with-output-file 'coke-indent-function 1)
(put 'with-output-to-file 'coke-indent-function 1)
(put 'with-output-to-port 'coke-indent-function 1)
(put 'call-with-values 'coke-indent-function 1)
(put 'dynamic-wind 'coke-indent-function 3)

(provide 'coke)
