{ import: st80 }
{ import: Random }

Alive : Object (age)
Alive initialize
[
	age := 1
]
Alive next
[
	age := age + 1.
	^ self
]

Dead  : Object ()

Alive nextWhen: aNumber
[
	aNumber = 2 ifTrue: [ ^ self next ].
	aNumber = 3 ifTrue: [ ^ self next ].
	^Dead.
]

Dead nextWhen: aNumber
[
	aNumber = 2 ifTrue: [ ^Dead ].
	aNumber = 3 ifTrue: [ ^Alive new ].
	^Dead.
]

Alive isAlive [ ^true ]
Dead isAlive [ ^false ]

Alive color
[
	| r b g |
	r := 0x80 + (age * 4) min: 0xff.
	g := 0xd0 - (age * 12) max: 0.
	b := 0xff - (age * 4) max: 0.
	^ (r bitShift: 16) + (g bitShift: 8) + b
]
Dead color [^0x000000]

Alive print [ age printString last put. ^1 ]
Dead  print [ $  put. ^1 ]

LifeMap : Object (map random)

LifeMap initialize
[
	super initialize.
	map ifNil: [self width: 20 height: 20].
	random := Random new.
]

LifeMap width: width height: height
[
	| col |
	map := Array new: width + 2.
	1 to: map size do: [ :x |
		col := Array new: height + 2.
		map at: x put: col.
		1 to: col size do: [ :y |
			col at: y put: Dead]].
]

LifeMap width [^ map size - 2]
LifeMap height [^ map first size - 2]

LifeMap at: x at: y
[
	^ (map at: x + 1) at: y + 1
]

LifeMap at: x at: y put: aBoolean
[
	^ (map at: x + 1) at: y + 1 put: aBoolean
]

LifeMap do: aBlock
[
	1 to: self width do: [ :x |
		1 to: self height do: [ :y | aBlock value: x value: y]]
]

LifeMap copy
[
	| new |
	new := self new.
	new width: self width height: self height.
	self do: [ :x :y |
		new at: x at: y put: (self at: x at: y)].
	^ new
]

LifeMap random
[
	self do: [ :x :y |
			random next > 0.5
				ifTrue: [self at: x at: y put: Alive new]
				ifFalse: [self at: x at: y put: Dead]]
]

LifeMap step
[
	| prev |
	prev := self copy.
	self do: [ :x :y |
		self at: x at: y
			put: ((self at: x at: y)
				nextWhen: (prev numAt: x at: y))]
]

LifeMap numAt: x at: y
[
	| num |
	num := 0.
	self ifAliveAt: x - 1 at: y - 1 do: [ num := num + 1 ].
	self ifAliveAt: x - 0 at: y - 1 do: [ num := num + 1 ].
	self ifAliveAt: x + 1 at: y - 1 do: [ num := num + 1 ].
	self ifAliveAt: x - 1 at: y + 0 do: [ num := num + 1 ].
	self ifAliveAt: x + 1 at: y + 0 do: [ num := num + 1 ].
	self ifAliveAt: x - 1 at: y + 1 do: [ num := num + 1 ].
	self ifAliveAt: x - 0 at: y + 1 do: [ num := num + 1 ].
	self ifAliveAt: x + 1 at: y + 1 do: [ num := num + 1 ].
	^num
]

LifeMap ifAliveAt: x at: y do: aBlock
[
	(self at: x at: y) isAlive
		ifTrue: [aBlock value]
]

LifeMap print
[
	map do: [ :each |		
		each print.
		$\n put.
	]
]
